#include "postgres.h"
#include "access/itup.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include "executor/spi.h"
#include "commands/trigger.h"
#include "nodes/pg_list.h"
#include "catalog/namespace.h"

#include "funcapi.h"
#include <fmgr.h>


#include "tsvector.h"
#include "query.h"
#include "ts_cfg.h"
#include "ts_stat.h"
#include "common.h"
#include "tsearchd/connection.h"
#include "tsearchd/tscommunication.h"

PG_FUNCTION_INFO_V1(tdindex);
Datum		tdindex(PG_FUNCTION_ARGS);

Datum
tdindex(PG_FUNCTION_ARGS) {
	int4 id = PG_GETARG_INT32(0);
	tsvector   *in = (tsvector *) PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
	TC_Connection conn;
	TSIndexData *data;
	TCMsg   *pmsg;
	u_int32_t rc = 0;

	TC_fillConnection(&conn, "127.0.0.1", TSEARCHD_PORT);
	TC_ServerInitConnect(&conn);

        conn.len = TCMSGHDRSZ + sizeof(TSIndexData) + in->len;
        conn.buf = palloc( conn.len );

        pmsg = (TCMsg*) conn.buf;
        pmsg->len = conn.len;
        pmsg->type = TSM_PUT;
	
	data = (TSIndexData*)pmsg->data;
	data->id = id;
	memcpy( data->data, in, in->len );

	if ( (rc=TC_Talk(&conn)) != CS_OK ) {
		TC_FreeConnection(&conn);
		 elog(ERROR,"Talk error to tsearchd: %d", rc);
	}
	if ( conn.ptr - conn.buf >= TCMSGHDRSZ + sizeof(u_int32_t) ) { 
		pmsg = (TCMsg*) conn.buf;
		rc = *(u_int32_t*)(pmsg->data);
	} else 
		rc = TSD_UNKOWNERROR;
	TC_FreeConnection(&conn);

	if ( rc != TSD_OK )
		elog(ERROR,"tsearchd returns %d", rc); 

	PG_FREE_IF_COPY(in, 1);
	PG_RETURN_VOID();
}

PG_FUNCTION_INFO_V1(tdflush);
Datum		tdflush(PG_FUNCTION_ARGS);

Datum
tdflush(PG_FUNCTION_ARGS) {
	TC_Connection conn;
	TCMsg   *pmsg;
	u_int32_t rc = 0;

	TC_fillConnection(&conn, "127.0.0.1", TSEARCHD_PORT);
	TC_ServerInitConnect(&conn);

	conn.len = TCMSGHDRSZ;
        conn.buf = palloc( conn.len );

        pmsg = (TCMsg*) conn.buf;
        pmsg->len = conn.len;
        pmsg->type = TSM_FLUSH;

	if ( (rc=TC_Talk(&conn)) != CS_OK ) {
		TC_FreeConnection(&conn);
		elog(ERROR,"Talk error to tsearchd: %d", rc);
	}
	if ( conn.ptr - conn.buf >= TCMSGHDRSZ + sizeof(u_int32_t) ) { 
		pmsg = (TCMsg*) conn.buf;
		rc = *(u_int32_t*)(pmsg->data);
	} else 
		rc = TSD_UNKOWNERROR;
	TC_FreeConnection(&conn);

	if ( rc != TSD_OK )
		elog(ERROR,"tsearchd returns %d", rc); 

	PG_RETURN_VOID();
}

typedef struct {
	TC_Connection conn;
	WordData  *data;
	TCMsg *pmsg;
	u_int32_t i;
} TSDStore;


static void
td_setup_first_call(FuncCallContext *funcctx, char* server, QUERYTYPE *tsquery) {
	MemoryContext     oldcontext;
	TSDStore  *store;
	TCMsg   *pmsg;
	TC_Connection *conn;

	oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
	store = (TSDStore*) palloc( sizeof(TSDStore) );
	store->data = NULL;
	store->i = 0;
	conn = &(store->conn);

	TC_fillConnection( conn, "127.0.0.1", TSEARCHD_PORT);
	TC_ServerInitConnect( conn );
	
	conn->len = TCMSGHDRSZ + tsquery->len;
	conn->buf = palloc( conn->len );
	
	pmsg = (TCMsg*) conn->buf;
	pmsg->len = conn->len;
	pmsg->type = TSM_QUERY;
	memcpy( pmsg->data, tsquery, tsquery->len );

	while( conn->state == CS_INPROCESS )
		TC_ServerConnect( conn, 100 );

	if ( conn->state != CS_CONNECTED ) {
		u_int32_t ret = conn->state;
		TC_FreeConnection( conn );
		elog(ERROR,"Connect error to tsearchd: %d", ret);
	}
	
	conn->ptr=conn->buf;
	conn->state = CS_SEND;
	while( conn->state != CS_FINISHSEND ) {
		while( !TC_ReadyIO( &conn, 1, 100) );
		TC_Send( conn );
	}

	if ( conn->state == CS_ERROR ) {
		u_int32_t ret = conn->state;
		TC_FreeConnection( conn );
		elog(ERROR,"Send error to tsearchd: %d", ret);
	}

	conn->state = CS_READ;	
	conn->ptr = conn->buf;
	
	funcctx->user_fctx = (void*)store;
	MemoryContextSwitchTo(oldcontext);
}	

static void
td_setup_second_call(TSDStore *store) {
	TC_Connection *conn =  &(store->conn);

	while( conn->state != CS_FINISHREAD ) {
		while( !TC_ReadyIO( &conn, 1, 100) );
		if ( TC_Read( conn ) == CS_ERROR ) {
			u_int32_t ret = conn->state;
			TC_FreeConnection( conn );
			elog(ERROR,"Read error to tsearchd: %d", ret);
		}
	}

	if ( conn->len < TCMSGHDRSZ + WDHRDRSZ ) {
		TC_FreeConnection( conn );
		elog(ERROR,"Unknown error on tsearchd");
	}

	store->pmsg = (TCMsg*)(conn->buf);
	conn->buf = NULL;
	TC_FreeConnection( conn );
	
	store->data = (WordData*)(store->pmsg->data);

	if ( store->data->status != TSD_OK )
		elog(ERROR,"tsearchd returns: %d", store->data->status); 
}

PG_FUNCTION_INFO_V1(tdsearch);
Datum tdsearch(PG_FUNCTION_ARGS);
Datum
tdsearch(PG_FUNCTION_ARGS) {
	FuncCallContext  *funcctx;
	TSDStore *store;

	if (SRF_IS_FIRSTCALL()) {
		QUERYTYPE *tsquery = (QUERYTYPE *) PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
		funcctx = SRF_FIRSTCALL_INIT();
		td_setup_first_call(funcctx, NULL, tsquery);
		PG_FREE_IF_COPY(tsquery, 0);
	}

	funcctx = SRF_PERCALL_SETUP();
	store = (TSDStore*) funcctx->user_fctx;

	if ( !store->data ) {
		td_setup_second_call(store);
		funcctx->max_calls = store->data->num; 
	}

	if ( store->i < store->data->num ) { 
		Datum res=Int32GetDatum( store->data->id[ store->i ] );
		store->i++;
		SRF_RETURN_NEXT(funcctx, res);
	}

	pfree( store->pmsg );
	pfree( store );

	SRF_RETURN_DONE(funcctx);
}

static bool
getTDStatus(TC_Connection *conn) {
	TCMsg   *pmsg;
	u_int32_t rc;

	if ( conn->state == CS_CONNECTED )
		return FALSE;
	else if ( conn->state != CS_FINISHSEND ) 
		elog(ERROR, "getTDStatus: Internal logic error, state: %d", conn->state);
		
	conn->state = CS_READ;
	conn->ptr = conn->buf;	
	while( conn->state != CS_FINISHREAD ) {
		while( !TC_ReadyIO( &conn, 1, 100) );
		if ( TC_Read(conn) == CS_ERROR ) {
			rc = conn->state;
			TC_FreeConnection(conn);
			elog(ERROR,"Read error from tsearchd: %d", rc);
		}
	}
	if ( conn->ptr - conn->buf >= TCMSGHDRSZ + sizeof(u_int32_t) ) {
		pmsg = (TCMsg*) conn->buf;
		rc = *(u_int32_t*)(pmsg->data);
	} else
		rc = TSD_UNKOWNERROR;

	if ( rc != TSD_OK ) {
		TC_FreeConnection(conn);
		elog(ERROR,"tsearchd returns %d at insert", rc);
	}
	return TRUE;
}

static void
TDSend(TC_Connection *conn) {
	u_int32_t rc;

	if ( !(conn->state == CS_CONNECTED || conn->state == CS_FINISHREAD) )
		elog(ERROR, "TDSend: Internal logic error, state: %d", conn->state);

	conn->state = CS_SEND;
	conn->ptr = conn->buf;

	while( conn->state != CS_FINISHSEND ) {
		while( !TC_ReadyIO( &conn, 1, 100) );
		if ( TC_Send(conn) == CS_ERROR ) {
			rc = conn->state;
			TC_FreeConnection(conn);
			elog(ERROR,"Send error to tsearchd: %d", rc);
		}
	}
}


PG_FUNCTION_INFO_V1(tdbulkindex);
Datum tdbulkindex(PG_FUNCTION_ARGS);
Datum
tdbulkindex(PG_FUNCTION_ARGS) {
	text *qt = PG_GETARG_TEXT_P(0);
	char       *query = text2char(qt);
	Portal          portal;
	void       *plan;
	Oid tiOid;
	int total=0;
	u_int32_t rc=CS_OK;
	TC_Connection	conn;
	TCMsg   *pmsg;


	TC_fillConnection(&conn, "127.0.0.1", TSEARCHD_PORT);
	TC_ServerInitConnect(&conn);
	conn.len = TCMSGHDRSZ + sizeof(TSIndexData);
	conn.buf = palloc( conn.len );

	PG_FREE_IF_COPY(qt, 0);

	SPI_connect();

	tiOid = get_ti_Oid();
	
	if ((plan = SPI_prepare(query, 0, NULL)) == NULL)
		elog(ERROR, "SPI_prepare('%s') returns NULL", query);

	if ((portal = SPI_cursor_open(NULL, plan, NULL, NULL, false)) == NULL)
		elog(ERROR, "SPI_cursor_open('%s') returns NULL", query);

	SPI_cursor_fetch(portal, true, 100);
	if (SPI_tuptable->tupdesc->natts != 2) {
		TC_FreeConnection(&conn);
		elog(ERROR, "number of fields doesn't equal to 1");
	}

	if (SPI_gettypeid(SPI_tuptable->tupdesc, 1) != INT4OID ) {
		TC_FreeConnection(&conn);
		elog(ERROR, "First column isn't of int4 type");
	}

	if (SPI_gettypeid(SPI_tuptable->tupdesc, 2) != tiOid) {
		TC_FreeConnection(&conn);
		elog(ERROR, "Second column isn't of tsvector type");
	}

	/* wait to connect */
	while( conn.state == CS_INPROCESS )
		TC_ServerConnect(&conn, 100);

	if ( conn.state != CS_CONNECTED ) {
		rc = conn.state;
		TC_FreeConnection(&conn);
		elog(ERROR, "Can't connect to tsearchd: %d", rc);
	}

	while (SPI_processed > 0) {
		int i;
		for (i = 0; i < SPI_processed; i++) {
			Datum id, tsv;
			bool   isnull;
			tsvector   *in;
			TSIndexData *data;

			id = SPI_getbinval(SPI_tuptable->vals[i], SPI_tuptable->tupdesc, 1, &isnull);
			if (isnull) continue; 
			tsv = SPI_getbinval(SPI_tuptable->vals[i], SPI_tuptable->tupdesc, 2, &isnull);
			if (isnull) continue; 
			in = (tsvector *) PG_DETOAST_DATUM(tsv);

			if ( getTDStatus(&conn) )
				total++;

			conn.len = TCMSGHDRSZ + sizeof(TSIndexData) + in->len;
			conn.buf = repalloc( conn.buf, conn.len );

			pmsg = (TCMsg*) conn.buf;
			pmsg->len = conn.len;
			pmsg->type = TSM_BULKPUT;

			data = (TSIndexData*)pmsg->data;
			data->id = DatumGetInt32(id);
			memcpy( data->data, in, in->len );

			TDSend(&conn);

			if ( in != (tsvector*)DatumGetPointer(tsv) )
				pfree(in);
		} 
		SPI_freetuptable(SPI_tuptable);
		SPI_cursor_fetch(portal, true, 100);
	}

	SPI_freetuptable(SPI_tuptable);
	SPI_cursor_close(portal);
	SPI_freeplan(plan);

	SPI_finish();

	if ( getTDStatus(&conn) )
		 total++;

	pmsg = (TCMsg*) conn.buf;
	pmsg->len = conn.len;
	pmsg->type = TSM_FLUSH;

	conn.state = CS_CONNECTED;
	conn.ptr = conn.buf;
	if ( (rc=TC_Talk(&conn)) != CS_OK ) {
		TC_FreeConnection(&conn);
		elog(ERROR,"Talk error to tsearchd: %d", rc);
	}

	if ( conn.ptr - conn.buf >= TCMSGHDRSZ + sizeof(u_int32_t) ) {
		pmsg = (TCMsg*) conn.buf;
		rc = *(u_int32_t*)(pmsg->data);
	} else
		rc = TSD_UNKOWNERROR;
	TC_FreeConnection(&conn);

	if ( rc != TSD_OK )
		elog(ERROR,"tsearchd returns %d at flush", rc);
	
	pfree(query);

	PG_RETURN_INT32(total);
}
