#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "tlog.h"
#include "tmalloc.h"

#include "tsearchd.h"


static WordData* 
findInCache(TDStorage *storage, char *word) {
	StackEntry	*ptr = storage->cache;

	while(ptr) {
		if ( strcmp( ptr->word, word ) == 0 )
			return ptr->data; 
		ptr=ptr->next;
	}
	return NULL;
}

static void
putInCache(TDStorage *storage, WordData *data, char *word) {
	StackEntry *s = storage->cache;

	while(s) {
		if ( s->data == data )
			return;
		s = s->next;
	} 

	s = (StackEntry*)t0malloc( sizeof(StackEntry) );
	s->next = storage->cache;
	storage->cache = s;
	s->word = word;
	s->data = data;
}

static void
freeChain(StackEntry *ptr, WordData *res) {
	StackEntry *tmp;
	while(ptr) {
		if ( !ptr->donotfree && ptr->data && res != ptr->data)
			tfree(ptr->data);
		tmp=ptr->next;
		tfree(ptr);
		ptr=tmp;
	}
}

static void
freeExec(TDStorage *storage, WordData *res) {
	freeChain(storage->cache, res); 
	storage->cache=NULL;
	freeChain(storage->stack, res); 
	storage->stack=NULL;
}


static void
replaceTop(TDStorage *storage, WordData* data) {
	StackEntry *ptr = storage->stack;

	storage->stack = ptr->next;
	if ( !ptr->donotfree )
		tfree( ptr->data );
	tfree(ptr);

	ptr = storage->stack;
	if ( !ptr->donotfree )
		tfree( ptr->data );
	ptr->data=data;
	ptr->donotfree=FALSE;
}

static void 
pushWord(TDStorage *storage, char *word) {
	StackEntry *s = (StackEntry*)t0malloc( sizeof(StackEntry) );

	s->data = findInCache(storage, word);
	if ( !s->data ) {
		s->data = TDGet(storage, word);
		if (!s->data) 
			s->data=(WordData*)t0malloc( WDHRDRSZ );
	} else
		s->donotfree=TRUE;

	s->next = storage->stack;
	storage->stack = s;
}

static void
andStack(TDStorage *storage) {
	StackEntry *opr1=storage->stack, *opr2;
	WordData *res;
	u_int32_t num;
	int32_t *r, *i1, *i2;

	assert( opr1 && opr1->next );
	opr2 = opr1->next;

	num = ( opr1->data->num > opr2->data->num ) ? opr2->data->num : opr1->data->num;
	res= (WordData*)tmalloc( WDHRDRSZ + num*sizeof(int32_t) );

	r = res->id;
	i1 = opr1->data->id;
	i2 = opr2->data->id;

	while( i1 - opr1->data->id < opr1->data->num && i2 - opr2->data->id < opr2->data->num ) {
		if ( *i1 == *i2 ) {
			*r = *i1;
			r++; i1++; i2++;
		} else if ( *i1 < *i2 ) 
			i1++;
		else
			i2++;
	} 

	res->num = r - res->id;

	replaceTop(storage, res); 
}

static void
orStack(TDStorage *storage) {
	StackEntry *opr1=storage->stack, *opr2;
	WordData *res;
	u_int32_t num;
	int32_t *r, *i1, *i2;

	assert( opr1 && opr1->next );
	opr2 = opr1->next;

	num = opr2->data->num + opr1->data->num;
	res= (WordData*)tmalloc( WDHRDRSZ + num*sizeof(int32_t) );

	r = res->id;
	i1 = opr1->data->id;
	i2 = opr2->data->id;

	while( i1 - opr1->data->id < opr1->data->num && i2 - opr2->data->id < opr2->data->num ) {
		if ( *i1 == *i2 ) {
			*r = *i1;
			r++; i1++; i2++;
		} else if ( *i1 < *i2 ) {
			*r = *i1; 
			i1++; r++;
		} else {
			*r = *i2;
			i2++; r++;
		}
	}

	while ( i1 - opr1->data->id < opr1->data->num ) {
		*r = *i1;
		i1++; r++;
	} 

	while ( i2 - opr2->data->id < opr2->data->num ) {
		*r = *i2;
		i2++; r++;
	} 

	res->num = r - res->id;

	replaceTop(storage, res); 
}

static void
notStack(TDStorage *storage) {
	StackEntry *opr2=storage->stack, *opr1;
	WordData *res;
	u_int32_t num;
	int32_t *r, *i1, *i2;

	assert( opr2 && opr2->next );
	opr1 = opr2->next;

	num = opr1->data->num;
	res= (WordData*)tmalloc( WDHRDRSZ + num*sizeof(int32_t) );

	r = res->id;
	i1 = opr1->data->id;
	i2 = opr2->data->id;

	while( i1 - opr1->data->id < opr1->data->num && i2 - opr2->data->id < opr2->data->num ) {
		if ( *i1 == *i2 ) {
			i1++; i2++;
		} else if ( *i1 < *i2 )  {
			*r=*i1;
			i1++; r++;
		} else 
			i2++;
	} 

	while ( i1 - opr1->data->id < opr1->data->num ) {
		*r = *i1;
		i1++; r++;
	} 

	res->num = r - res->id;

	replaceTop(storage, res); 
}

WordData*
TDExecute(TDStorage *storage, TDNode *root) {
	TDNode *node = root;
	WordData *res=NULL;

	while( node ) {
		switch( node->type ) {
			case TD_VAL:
				pushWord(storage, node->word);
				if (node->cancache) {
					putInCache(storage, storage->stack->data, node->word);
					storage->stack->donotfree = TRUE;
				}
				break;
			case TD_AND:
				andStack(storage);
				break;
			case TD_OR:
				orStack(storage);
				break;
			case TD_MINES:
				notStack(storage);
				break;
			default:
				tlog(TL_ALARM, "Unkown operation: %d", node->type);
				freeExec( storage, NULL );
				return NULL;
		}

		if ( node->isfinita && storage->stack->data->num == 0 ) {
			freeExec( storage, NULL );
			return NULL;
		}
		node = node->next; 
	}

	assert( storage->stack->next == NULL );

	res = storage->stack->data;
	freeExec( storage, res );

	return res;	
}

