#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "tlog.h"
#include "tmalloc.h"

#include "tsearchd.h"

/* encodes integer n into buffer (buffer length must be at least 5 bytes),
 returns number of bytes used in the buffer */
static int
vby_encode(char *buf, u_int32_t n) {
	int nb;

	nb = 0;
	while (n >= 128) {
		buf[nb++] = (n & 0x7f) | 0x80;
		n >>= 7;
	}
	buf[nb++] = n;
	return nb;
}

/* decodes integer n from buffer (buffer length must be at least 5 bytes),
 returns number of bytes used from the buffer */
static int
vby_decode (char *buf, u_int32_t *n) {
	int nb;
	u_int32_t count, get;

	*n = 0;
	nb = 0;
	count = 0;

	while (TRUE) {
		if (((get = buf[nb++]) & 0x80) == 0x80) {
			/* For each time we get a group of 7 bits, need to left-shift the
			   latest group by an additional 7 places, since the groups were
			   effectively stored in reverse order.*/
			*n |= ((get & 0x7f) << count);
			count += 7;
		} else if (nb > (sizeof(u_int32_t))) {
			if (nb == ((sizeof(u_int32_t) + 1))
				 && (get < (1 << (sizeof(u_int32_t) + 1)))) {
				/* its a large, but valid number */
				*n |= get << count;
				return nb;
			} else {
				/* we've overflowed, return error */
				return 0;
			}
		} else {
			/* Now get the final 7 bits, which need to be left-shifted by a
			   factor of 7. */
			*n |= get << count;
			return nb;
		}
	}
	return 0;
}

TDStorage* 
TDOpenStorage(char *file) {
	TDStorage *storage=(TDStorage*)t0malloc(sizeof(TDStorage));
	
	if ( (storage->db=dbopen(file, O_CREAT|O_RDWR, 0666, DB_BTREE, NULL)) == NULL ) 
		tlog( TL_CRIT|TL_EXIT, "dbopen failed with status %d (%s)", errno, strerror(errno) );

	storage->maxallocated = 4*1024*1024; /* Mb */

	return storage;
}

static AccEntry *
EntryAlloc( TDStorage *storage, AccEntry *entry, u_int32_t n) {
	if ( entry ) {
		entry = ( AccEntry * ) trealloc( entry, ACCEHRDRSZ + sizeof(int32_t) * n );
		storage->allocated -= ACCEHRDRSZ + sizeof(int32_t) * entry->len;
	} else 
		entry = ( AccEntry * ) tmalloc( ACCEHRDRSZ + sizeof(int32_t) * n );

	entry->len = n;
	storage->allocated += ACCEHRDRSZ + sizeof(int32_t) * entry->len;

	return entry;	
} 

static void
EntryArrayAlloc( TDStorage *storage, u_int32_t n) {
	if ( storage->entries ) {
		storage->entries = (AccEntry**) trealloc( storage->entries, sizeof(AccEntry*) * n );
		storage->allocated -= sizeof(AccEntry*) * storage->lenentries;
	} else
		storage->entries = (AccEntry**) tmalloc( sizeof(AccEntry*) * n );

	storage->lenentries = n;
	storage->allocated += sizeof(AccEntry*) * storage->lenentries;
}

void 
TDAppend( TDStorage *storage, char *word, int32_t id) {
	AccEntry *ptr=NULL;

	if ( !storage->entries ) {
		EntryArrayAlloc(storage, 128);
		*(storage->entries) = ptr = EntryAlloc(storage, NULL, 4);
		ptr->id[0] = id;
		ptr->num = storage->numentries = 1;
		ptr->word=tstrdup( word );
		storage->allocated += (u_int32_t) strlen(word);
	} else {
		AccEntry **StopLow  = storage->entries;
		AccEntry **StopHigh = storage->entries + storage->numentries;
		AccEntry **StopMiddle;
		int rc;

		/* Loop invariant: StopLow <= StopMiddle < StopHigh */
		while (StopLow < StopHigh) {
			StopMiddle = StopLow + ((StopHigh - StopLow) >> 1);
			rc = strcmp( (*StopMiddle)->word, word );
			if ( rc==0 ) {
				ptr = *StopMiddle;
				if ( ptr->num >= ptr->len ) {
					*StopMiddle = EntryAlloc(storage, ptr, 2 * ptr->len);
					ptr = *StopMiddle;
				} 
				ptr->id[ ptr->num ] = id;
				ptr->num++;
				return;
			} else if ( rc<0 )
				StopLow = StopMiddle + 1;
			else 
				StopHigh = StopMiddle;
		}
	
		if (!ptr && StopLow>=StopHigh) {
			if ( storage->numentries >= storage->lenentries ) {
				int diff = StopHigh - storage->entries;
				EntryArrayAlloc(storage, 2 * storage->lenentries);
				StopHigh = storage->entries + diff;
			}
			if ( StopHigh!= storage->entries + storage->numentries) 
				memmove( StopHigh+1, StopHigh, 
					sizeof(AccEntry*) * (storage->entries + storage->numentries - StopHigh) );
			ptr = *StopHigh = EntryAlloc(storage, NULL, 4);
			ptr->id[0] = id;
			ptr->num = 1;
			ptr->word=tstrdup( word );
			storage->allocated += (u_int32_t) strlen(word);
			storage->numentries++;  
		} 
	}	
}

WordData*
TDGet(TDStorage *storage, char *word) {
	DBT k, v;
	WordData *res=NULL;
	int rc;
	char *ptr;
	u_int32_t num,i=0;
	
	memset (&k, 0, sizeof (DBT));
	memset (&v, 0, sizeof (DBT));

	k.data = word;
	k.size = strlen(word);

	rc = (storage->db)->get((storage->db), &k, &v, 0);
	switch(rc) {
		case 1:
			tlog(TL_DEBUG,"TDGet: word %s notfound", word); 
			break;
		case 0:
			ptr = (char*)v.data;
			ptr += vby_decode(ptr, &num);
			res=(WordData*)tmalloc( WDHRDRSZ + num*sizeof(int32_t) );
			res->status=0;
			res->num=num;
			while ( ptr - ((char*)v.data) < v.size ) {
				ptr += vby_decode(ptr, (u_int32_t*)( res->id + i ) );
				if ( i )
					res->id[i] += res->id[i-1] + 1;
				i++;
			}
			tlog(TL_DEBUG,"TDGet: word %s Nresult:%d size:%d bytes", word, res->num, v.size); 
			break;
		case -1:
		default:
			tlog(TL_CRIT|TL_EXIT, "dbget failed %d (%s)", errno, strerror(errno));
	}
	return res;	
}

void 
TDPut(TDStorage *storage, char *word, WordData *data) {
	DBT k, v;
	u_int32_t i;
	static u_int32_t tmplen=0;
	static char *tmpbuf = NULL;
	char *ptr;

	memset (&k, 0, sizeof (DBT));
	memset (&v, 0, sizeof (DBT));
	
	k.data = word;
	k.size = strlen(word);

	if ( tmpbuf==NULL || ( sizeof(int32_t) * (data->num + 1) ) * 5 / 4 + 4 > tmplen ) {
		tmplen = ( sizeof(int32_t) * (data->num+1) ) * 5 / 4 + 4;
		tmpbuf = (WordData*) (tmpbuf) ? trealloc(tmpbuf, tmplen) : tmalloc(tmplen);
	}

	ptr = tmpbuf;
	ptr += vby_encode( ptr, data->num );
	tlog(TL_DEBUG,"PUT %d", *tmpbuf);
	for(i=0;i<data->num;i++)
		if ( i==0 ) 
			ptr += vby_encode( ptr, (u_int32_t)(data->id[0]) );
		else 
			ptr += vby_encode( ptr, (u_int32_t)( data->id[i] -  data->id[i-1] - 1 ) ); 

	v.data = (void*)tmpbuf;
	v.size = (int)( ptr-tmpbuf );

	tlog(TL_DEBUG,"TDPut: word %s Nresult:%d size:%d bytes", word, data->num, v.size); 
	
	if ( (storage->db)->put((storage->db), &k, &v, 0) != 0 ) 
		tlog(TL_CRIT|TL_EXIT, "dbput failed %d (%s)", errno, strerror(errno));
}

static int
idCMP(const void *a, const void *b) {
	if ( *(int32_t*)a == *(int32_t*)b )
		return 0;
	return ( *(int32_t*)a > *(int32_t*)b ) ? 1 : -1;
	
}

static u_int32_t
uniqueint(int32_t *a, u_int32_t l)
{
        int32_t       *ptr,
                           *res;

        if (l <  2)
                return l;

        ptr = res = a;

        qsort((void *) a, l, sizeof(int32_t), idCMP);

        while (ptr - a < l)
                if (*ptr != *res)
                        *(++res) = *ptr++;
                else
                        ptr++;
        return (u_int32_t)(res + 1 - a);
}


void
TDFlushStorage(TDStorage *storage) {
	u_int32_t i;

	if ( storage->numentries == 0 )
		return;

	tlog(TL_DEBUG,"TDFlushStorage: allocated:%d maxallocated:%d",
		storage->allocated, storage->maxallocated);
 
	for(i=0; i<storage->numentries; i++ ) {
		AccEntry *entry = storage->entries[i];
		WordData *newdata = (WordData*) ( ((char*)entry) + ACCEWDHRDRSZ );
		WordData *olddata = TDGet(storage, entry->word);

		if ( olddata ) {
			newdata = (WordData*)tmalloc( WDHRDRSZ + sizeof(int32_t) * (entry->num + olddata->num) );
			newdata->num = entry->num + olddata->num;
			memcpy( newdata->id, olddata->id, sizeof(int32_t)*olddata->num );
			memcpy( newdata->id + olddata->num, entry->id, sizeof(int32_t)*entry->num );
		}

		newdata->num = uniqueint(newdata->id, newdata->num); 

		TDPut( storage, entry->word, newdata );
		if ( newdata != (WordData*) ( ((char*)entry) + ACCEWDHRDRSZ ) )
			tfree(newdata);
		if (olddata)
			tfree(olddata);
		tfree(entry->word);
		tfree(entry);
	}

	tfree( storage->entries );
	storage->entries = NULL;	
	storage->lenentries = storage->numentries = storage->allocated = 0;

	if ( (storage->db)->sync((storage->db), 0) != 0 ) 
		tlog(TL_CRIT|TL_EXIT, "dbsync failed %d (%s)", errno, strerror(errno));
}

void 
TDCloseStorage(TDStorage *storage) {
	TDFlushStorage(storage);

	if ( (storage->db)->close(storage->db) < 0 ) 
		tlog( TL_CRIT|TL_EXIT, "dbclose failed with status %d (%s)", errno, strerror(errno) );

	tfree(storage);	
}


